"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const request = require("request");
const url = require("url");
const _l = require("lodash");
const firebase = require("firebase");
const credentials_1 = require("./credentials");
const utils_1 = require("./utils");
exports.COMPILESERVER = process.env['PAGEDRAW_COMPILESERVER'] || 'https://happy-unicorns.herokuapp.com';
exports.METASERVER = process.env['PAGEDRAW_METASERVER'] || 'https://pagedraw.io/';
exports.DOCSERVER = process.env['PAGEDRAW_DOCSERVER'] || 'https://pagedraw.firebaseio.com/';
const docserver_firebase = firebase.initializeApp({ databaseURL: exports.DOCSERVER });
function firebaseDB() {
    return docserver_firebase.database();
}
exports.firebaseDB = firebaseDB;
exports.netrc_entry = process.env['NETRC_ENTRY'] || 'pagedraw.io';
exports.auth = new credentials_1.CredentialStore(exports.netrc_entry);
function watchDoc(docserver_id, callback) {
    const ref = firebaseDB().ref(`pages/${docserver_id}`);
    const watch_id = ref.on('value', ((page) => { callback(); }), ((error) => { throw error; }));
    return () => { ref.off('value', watch_id); };
}
exports.watchDoc = watchDoc;
function getApp(app_name) {
    return new Promise((accept, reject) => {
        request.get(url.resolve(exports.METASERVER, `api/v1/cli/apps/${app_name}?id=${exports.auth.credentials.id}&auth_token=${exports.auth.credentials.auth_token}`), (err, resp, body) => {
            if (resp.statusCode == 404) {
                return reject(new utils_1.CLIPrintableError(`Unable to fetch data from Pagedraw API. Are you sure you have access to the app ${app_name}? Try running pagedraw login`));
            }
            if (err || resp.statusCode != 200) {
                return reject(new utils_1.CLIPrintableError('Unable to fetch data from the Pagedraw API. Are you connected to the internet?'));
            }
            let appData;
            try {
                appData = JSON.parse(body);
            }
            catch (err) {
                return reject(new utils_1.CLIPrintableError('Pagedraw API returned bad JSON.'));
            }
            accept(appData);
        });
    });
}
exports.getApp = getApp;
function metaserver_authed_rpc(endpoint, data) {
    request.post({
        uri: url.resolve(exports.METASERVER, endpoint),
        json: true,
        body: _l.extend({}, data, { id: exports.auth.credentials.id, auth_token: exports.auth.credentials.auth_token })
    });
}
exports.metaserver_authed_rpc = metaserver_authed_rpc;
function compileFromDocserverId(docserver_id) {
    return __awaiter(this, void 0, void 0, function* () {
        let [err, resp, body] = yield new Promise((accept, reject) => {
            request.post({
                uri: url.resolve(exports.COMPILESERVER, `v1/compile/${docserver_id}`),
                json: true,
                body: { client: 'cli', user_info: { id: exports.auth.credentials.id } }
            }, (err, resp, body) => accept([err, resp, body]));
        });
        if (err) {
            throw err;
        }
        if (resp.statusCode !== 200) {
            return {
                files: null,
                messages: [{
                        level: "error",
                        message: "Error compiling doc. This is an internal pagedraw error.  If it persists, please contact team@pagedraw.io"
                    }]
            };
        }
        if (!_l.isArray(body)) {
            return {
                files: null,
                messages: [{
                        level: "error",
                        message: "Error communicating with server.  Try upgrading this CLI.  If this persists, please contact team@pagedraw.io"
                    }]
            };
        }
        let messages = [];
        let broken_entries;
        [body, broken_entries] = _l.partition(body, (entry) => _l.isString(entry.filePath) && _l.isString(entry.contents));
        if (!_l.isEmpty(broken_entries)) {
            messages.push({
                level: "error",
                message: `Error communicating with server for ${broken_entries.length} files.  Please tell us at team@pagedraw.io`
            });
            if (!_l.isEmpty(body)) {
                return { files: null, messages };
            }
        }
        let missing_filepath;
        [missing_filepath, body] = _l.partition(body, (entry) => _l.isEmpty(entry.filePath));
        if (!_l.isEmpty(missing_filepath)) {
            messages.push({
                level: "error",
                message: `Up to ${missing_filepath.length} components are missing a file path and were not synced.`,
            });
            if (!_l.isEmpty(body)) {
                return { files: null, messages };
            }
        }
        body = _l.map(_l.groupBy(body, 'filePath'), (entries_with_same_filePath, filePath) => {
            if (entries_with_same_filePath.length > 1) {
                messages.push({
                    level: "error",
                    message: `multiple components with the same file path: ${filePath}. Only syncing one of them...`,
                    filePath
                });
            }
            return entries_with_same_filePath[0];
        });
        messages.concat(_l.flatMap(body, (entry) => entry.errors.map((error) => { return { level: 'error', message: error.message, filePath: entry.filePath }; })), _l.flatMap(body, (entry) => entry.warnings.map((warning) => { return { level: 'warning', message: warning.message, filePath: entry.filePath }; })));
        return { files: body, messages };
    });
}
exports.compileFromDocserverId = compileFromDocserverId;
//# sourceMappingURL=api.js.map