"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const http = require("http");
const findup = require("findup");
const express = require("express");
const resolveFrom = require("resolve-from");
const semver = require("semver");
const cors = require("cors");
const _l = require("lodash");
const MemoryFS = require("memory-fs");
const request = require("request-promise-native");
const bodyParser = require("body-parser");
const md5 = require("blueimp-md5");
const mime = require("mime-types");
const fs = require("fs");
const load_config_1 = require("./load_config");
const hash_string = (str) => md5(str);
function get_project_info() {
    return __awaiter(this, void 0, void 0, function* () {
        const detected_project_info = yield Promise.all(['package.json', 'bower.json'].map((filename) => __awaiter(this, void 0, void 0, function* () {
            try {
                const search_result = yield find_in_dir_or_ancestor(process.cwd(), filename);
                if (search_result.found) {
                    const package_file = search_result.found ? JSON.parse(fs.readFileSync(search_result.path, 'utf8')) : {};
                    const root_path = search_result.found ? path.dirname(search_result.path) : undefined;
                    return { project_type: detect_framework(package_file), root_path };
                }
                else {
                    return { project_type: 'UNKNOWN', root_path: undefined };
                }
            }
            catch (e) {
                return { project_type: 'UNKNOWN', root_path: undefined };
            }
        })));
        let info = detected_project_info.reduce((inferred, other) => (inferred.project_type === 'UNKNOWN' ? other : inferred), { project_type: 'UNKNOWN', root_path: undefined });
        if (info === undefined) {
            return { project_type: 'UNKNOWN', root_path: process.cwd() };
        }
        return {
            project_type: info.project_type,
            root_path: info.root_path || process.cwd()
        };
    });
}
function start_dev_server(port, config_file, override, specs_file) {
    return __awaiter(this, void 0, void 0, function* () {
        process.env.NODE_ENV = "development";
        const project_info = yield get_project_info();
        process.chdir(project_info.root_path);
        const compiler_data = yield make_external_code_compiler({
            host: `http://localhost:${port}`,
            static_id: '__dev',
            specs_file,
            config_file,
            override
        }, make_dev_config);
        if (compiler_data === undefined) {
            throw new Error("Could not start server");
        }
        const { compiler, dependencies, config } = compiler_data;
        const devMiddlewareOptions = {
            logLevel: 'error',
            noInfo: true,
            publicPath: config.output.publicPath,
            allowedHosts: [
                'localhost',
                '.pagedraw.io',
            ],
            compress: true,
            stats: 'errors-only',
        };
        let webpackResolve = (_) => { };
        let webpackReject = (_) => { };
        const webpackValid = new Promise((resolve, reject) => {
            webpackResolve = resolve;
            webpackReject = reject;
        });
        const router = express.Router();
        const dev_middleware_instance = dependencies.devMiddleware(compiler, devMiddlewareOptions);
        router.use(dev_middleware_instance);
        router.use(dependencies.hotMiddleware(compiler));
        dev_middleware_instance.waitUntilValid(stats => {
            if (stats.toJson().errors.length > 0) {
                webpackReject(stats);
            }
            else {
                webpackResolve(stats);
            }
        });
        const app = express();
        app.use(cors());
        app.use(router);
        app.use(bodyParser.json());
        app.post('/exit_dev', (req, res) => {
            process.env.NODE_ENV = "production";
            const static_id = req.body.static_id;
            const host = req.body.host;
            const metaserver = req.body.metaserver;
            if (typeof static_id !== 'string' || typeof host !== 'string' || typeof metaserver !== 'string') {
                res.send(JSON.stringify({
                    status: 'internal-err'
                }));
                return;
            }
            make_prod_bundle(host, static_id, config_file, override, specs_file).then((data) => {
                process.env.NODE_ENV = "development";
                if (data.user_errors) {
                    res.send(JSON.stringify({
                        status: 'user-err'
                    }));
                    return;
                }
                if (data.internal_errors) {
                    res.send(JSON.stringify({
                        status: 'internal-err'
                    }));
                    return;
                }
                if (_l.isEmpty(data.bundle)) {
                    console.error('Error: prod bundle cannot be empty');
                    res.send(JSON.stringify({
                        status: 'internal-err'
                    }));
                    return;
                }
                console.log("production bundle done. Starting upload...");
                return upload_static_files(metaserver, static_id, data.static_files)
                    .then(() => upload_code(metaserver, static_id, data.bundle))
                    .then(hash => {
                    console.log("Upload successful.");
                    res.send(JSON.stringify({
                        status: 'ok', id: hash,
                    }));
                }).catch((e) => {
                    res.send(JSON.stringify({
                        status: 'net-err'
                    }));
                });
            });
        });
        app.get('/are-you-alive', (req, res) => {
            res.send('yes');
        });
        const server = http.createServer(app);
        let serverResolve = () => { };
        let serverReject = (_) => { };
        const serverListening = new Promise((resolve, reject) => {
            serverResolve = resolve;
            serverReject = reject;
        });
        server.listen(port, error => {
            if (error) {
                serverReject(error);
            }
            else {
                serverResolve();
            }
        });
        try {
            yield Promise.all([webpackValid, serverListening]);
            console.log(`server started on http://localhost:${port}`);
        }
        catch (e) {
        }
    });
}
exports.start_dev_server = start_dev_server;
;
const make_config = (options, load_config) => {
    const base_config = load_config();
    let config;
    if (options.config_file) {
        const config_path = path.resolve(options.config_file);
        const user_config_search_result = load_config_1.try_to_load_user_config(config_path);
        if (!user_config_search_result.found) {
            console.error(`ERROR: unable to load custom configuration at ${config_path}`);
            return;
        }
        if (typeof user_config_search_result.config !== 'object') {
            console.error(`ERROR: Webpack config loaded from ${config_path} is not an object.`);
            console.error("Pagedraw does not yet support Webpack configurations as functions.");
            return;
        }
        if (options.override) {
            config = load_config_1.override_merge_configs(base_config, user_config_search_result.config);
        }
        else {
            config = load_config_1.merge_configs(base_config, user_config_search_result.config);
        }
    }
    else {
        config = base_config;
    }
    return config;
};
const make_dev_config = (version, options, dependencies) => make_config(options, () => load_config_1.load_dev_config(version, dependencies, path.resolve(options.specs_file), options.host));
const make_prod_config = (version, options, dependencies) => make_config(options, () => load_config_1.load_prod_config(version, path.resolve(options.specs_file), options.host, options.static_id));
const make_external_code_compiler = (options, make_config) => __awaiter(this, void 0, void 0, function* () {
    let webpack_path, webpack_deps, webpack_version;
    try {
        webpack_path = resolveFrom(process.cwd(), 'webpack');
    }
    catch (e) {
        webpack_path = require.resolve('webpack');
    }
    const package_search_result = yield find_in_dir_or_ancestor(path.dirname(webpack_path), 'package.json');
    if (!package_search_result.found) {
        console.warn("Could not deduce Webpack version. Assuming Webpack 4.x.");
        webpack_deps = require('pagedraw-cli-webpack4');
    }
    else {
        const package_json = JSON.parse(fs.readFileSync(package_search_result.path, 'utf8'));
        webpack_version = semver.major(package_json.version);
        switch (webpack_version) {
            case 3:
                webpack_deps = require('pagedraw-cli-webpack3');
                break;
            case 4:
                webpack_deps = require('pagedraw-cli-webpack4');
                break;
            default:
                console.warn(`Cannot recognize Webpack version ${package_json.version}. Falling back to Webpack 4.`);
                webpack_deps = require('pagedraw-cli-webpack4');
        }
    }
    const config = make_config(webpack_version, options, webpack_deps);
    if (config === undefined) {
        return;
    }
    const webpack = require(webpack_path);
    const compiler = webpack(config);
    return { compiler, dependencies: webpack_deps, config };
});
const upload_static_files = (metaserver, prefix, static_files) => {
    return Promise.all(Object.keys(static_files).map((name) => {
        const filetype = mime.lookup(name) || 'binary/octet-stream';
        return request({
            uri: `${metaserver}/sign/external_code`,
            qs: {
                filepath: path.join(prefix, name),
                filetype,
            },
        }).then((res) => {
            const data = JSON.parse(res);
            return request({
                uri: data.upload_url,
                method: 'PUT',
                body: static_files[name],
                headers: {
                    'Content-Type': filetype,
                },
            });
        }).catch((e) => {
            console.log("Metaserver returned error:");
            console.log(e);
        });
    }));
};
const upload_code = (metaserver, static_id, bundle) => {
    const filename = `${static_id}/${static_id}`;
    return request({
        uri: `${metaserver}/sign/external_code`,
        qs: {
            filepath: filename,
            filetype: 'application/javascript'
        },
    }).then((res) => {
        const data = JSON.parse(res);
        return request({
            uri: data.upload_url,
            method: 'PUT',
            body: bundle,
            headers: {
                'Content-Type': 'application/javascript',
            },
        });
    }).then(() => filename);
};
const webpack_compile = (compiler) => new Promise((resolve, reject) => {
    compiler.run((err, stats) => {
        if (err) {
            return resolve({ webpack_errors: [err.toString()] });
        }
        if (stats.hasErrors()) {
            console.error('Compilation error:');
            console.error(stats.toJson().errors.map(err => err.toString()));
            return resolve({ user_errors: stats.toJson().errors.map(err => err.toString()) });
        }
        return resolve({});
    });
});
function make_prod_bundle(host, static_id, config_file, override, specs_file) {
    return __awaiter(this, void 0, void 0, function* () {
        process.env.NODE_ENV = "production";
        const project_info = yield get_project_info();
        const base_path = project_info.root_path;
        process.chdir(base_path);
        const compiler_data = yield make_external_code_compiler({
            host,
            static_id,
            specs_file,
            config_file,
            override
        }, make_prod_config);
        if (compiler_data === undefined) {
            throw new Error("Could not start server");
        }
        const { compiler, dependencies, config } = compiler_data;
        const output_filename = 'bundle.js';
        const output = new MemoryFS();
        compiler.outputFileSystem = output;
        try {
            const compiler_output = yield webpack_compile(compiler);
            if (compiler_output.webpack_errors) {
                return { internal_errors: compiler_output.webpack_errors };
            }
            if (compiler_output.user_errors) {
                return { user_errors: compiler_output.user_errors };
            }
            const bundle = output.readFileSync(path.resolve(output_filename), 'utf8');
            const static_file_data = read_files_in_dir(output, process.cwd(), [output_filename]);
            const static_files = _l.fromPairs(static_file_data);
            return { bundle, static_files };
        }
        catch (e) {
            console.error(e);
            return { internal_errors: [e] };
        }
    });
}
exports.make_prod_bundle = make_prod_bundle;
;
const read_files_in_dir = (fs, root, skip) => {
    const _read_files_in_dir = (fs, root, current, skip) => {
        const filenames = fs.readdirSync(current);
        const children = filenames.map((filename) => {
            if (!skip.includes(filename)) {
                const filepath = path.join(current, filename);
                if (fs.statSync(filepath).isDirectory()) {
                    return _read_files_in_dir(fs, root, filepath, skip);
                }
                else {
                    const relative_path = path.relative(root, filepath);
                    return [[relative_path, fs.readFileSync(filepath)]];
                }
            }
        });
        return _l.flatten(_l.compact(children));
    };
    return _read_files_in_dir(fs, root, root, skip);
};
const detect_framework = (package_file) => {
    const has_dependency = (package_file, property) => ((package_file.dependencies && package_file.dependencies[property])
        || (package_file.devDependencies && package_file.devDependencies[property]));
    const has_peer_dependency = (package_file, property) => ((package_file.peerDependencies && package_file.peerDependencies[property]));
    if (has_dependency(package_file, 'react-scripts')) {
        return 'REACT-SCRIPTS';
    }
    if (has_dependency(package_file, 'react') && has_dependency(package_file, 'webpack')) {
        return 'WEBPACK-REACT';
    }
    if (has_dependency(package_file, 'react')) {
        return 'REACT';
    }
    if (has_peer_dependency(package_file, 'react')) {
        return 'USES-REACT';
    }
    return 'UNKNOWN';
};
const find_in_dir_or_ancestor = (base_dir, filename) => new Promise((resolve, reject) => {
    const immediate_path = path.resolve(base_dir, filename);
    if (fs.existsSync(immediate_path)) {
        return resolve({ path: immediate_path, found: true });
    }
    findup(base_dir, filename, (err, dir) => {
        if (err) {
            return resolve({ found: false });
        }
        return resolve({ path: path.join(dir, filename), found: true });
    });
});
//# sourceMappingURL=ext-dev-server.js.map