#!/usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
require('dotenv').config();
const _l = require("lodash");
const program = require("commander");
const clear = require("cli-clear");
const srs = require("secure-random-string");
const url = require("url");
const request = require("request-promise-native");
const open = require("open");
const fs = require("fs");
const path = require("path");
const pdAPI = require("./api");
const pdConfig = require("./config");
const utils_1 = require("./utils");
const pdSyncer = require("./pd-syncer");
const keep_up_to_date_1 = require("./keep_up_to_date");
const ext_dev_server_1 = require("./ext-dev-server");
const pkgJson = require('../package.json');
program
    .version(pkgJson.version, "-v, --version")
    .usage('<command>');
function print_cli_errors(main) {
    return __awaiter(this, void 0, void 0, function* () {
        try {
            yield main();
        }
        catch (err) {
            if (err instanceof utils_1.CLIPrintableError) {
                utils_1.abort(err.message);
            }
            else if (process.env["ENVIRONMENT"] === "development") {
                console.error(err);
                process.exit(1);
            }
            else {
                utils_1.abort("Error: CLI crashed");
            }
        }
    });
}
program
    .command('login')
    .description('Authenticate to gain access to the Pagedraw API.')
    .action((env, options) => {
    keep_up_to_date_1.enforceClientUpToDate(() => {
        console.log('Logging into Pagedraw');
        const local_token = srs({ length: 32 });
        const signin_url = url.resolve(pdAPI.METASERVER, `api/v1/cli/authenticate/${local_token}`);
        open(signin_url);
        console.log('Your browser has been open to visit:');
        console.log('    ' + signin_url);
        console.log('Waiting for authentication...');
        utils_1.poll_with_max_retries(500, 600, ((retry) => {
            request.get(url.resolve(pdAPI.METASERVER, `api/v1/cli/get_auth_token/${local_token}`), (err, resp, body) => {
                if (!(resp && resp.statusCode == 200)) {
                    return retry();
                }
                const { id, auth_token, email } = JSON.parse(body);
                pdAPI.auth.persist({ id: String(id), auth_token });
                console.log(`Authentication successful. You are now logged in as ${email}. ` +
                    `If this is the wrong account, login to your correct gmail account by running \'pagedraw login\` again.`);
                pdAPI.metaserver_authed_rpc('api/v1/cli/ran_pagedraw_login', {});
            }).catch((e) => {
                if (e.statusCode !== 404) {
                    console.error("There was an error while attempting authentication:");
                    console.error(e.message);
                }
            });
        }), (() => {
            utils_1.abort('Authentication timed out.');
        }));
    });
});
program
    .command('pull [docs_to_fetch...]')
    .description('Compile remote Pagedraw docs and pulls them into your local file system, in the path specified by the doc\'s file_path.')
    .action((docs_to_fetch) => {
    keep_up_to_date_1.enforceClientUpToDate(() => {
        print_cli_errors(() => __awaiter(this, void 0, void 0, function* () {
            let pd_config = yield pdConfig.loadPagedrawConfig();
            let appData = yield pdAPI.getApp(pd_config.app);
            console.log(`Pulling docs from app ${appData.name}...\n`);
            const docs = _l.isEmpty(docs_to_fetch)
                ? appData.pages
                : appData.pages.filter((doc) => docs_to_fetch.includes(doc.id.toString()) || docs_to_fetch.includes(doc.url));
            pdAPI.metaserver_authed_rpc('api/v1/cli/ran_pagedraw_pull', { app: appData, docs });
            yield Promise.all(docs.map((doc) => __awaiter(this, void 0, void 0, function* () {
                return pdSyncer.pullPagedrawDoc(doc, [doc], pd_config.managed_folders || []);
            })));
        }));
    });
});
program
    .command('sync [docs_to_fetch...]')
    .description('Compile remote Pagedraw docs and continuously sync them into your local file system, in the path specified by each doc\'s file_path.')
    .action(function (docs_to_fetch) {
    keep_up_to_date_1.long_term_enforce_client_up_to_date(() => {
        print_cli_errors(() => __awaiter(this, void 0, void 0, function* () {
            let pd_config = yield pdConfig.loadPagedrawConfig();
            let appData = yield pdAPI.getApp(pd_config.app);
            clear();
            console.log(`Syncing docs from app ${appData.name}.\nHit Ctrl + C to exit...`);
            const docs = _l.isEmpty(docs_to_fetch)
                ? appData.pages
                : appData.pages.filter((doc) => docs_to_fetch.includes(doc.id.toString()) || docs_to_fetch.includes(doc.url));
            pdAPI.metaserver_authed_rpc('api/v1/cli/ran_pagedraw_sync', { app: appData, docs: docs });
            yield Promise.all(docs.map((doc) => __awaiter(this, void 0, void 0, function* () {
                return pdAPI.watchDoc(doc.docserver_id, () => __awaiter(this, void 0, void 0, function* () {
                    yield pdSyncer.pullPagedrawDoc(doc, [doc], pd_config.managed_folders || []);
                    console.log('');
                }));
            })));
        }));
    });
});
program
    .command('develop [specsFile]')
    .description('Start a development server and develop libraries for Pagedraw')
    .option('-c, --config <config>', 'use custom webpack configuration with default loaders')
    .option('-r, --override-config <config>', 'use fully custom webpack configuration')
    .action((specsFile, cmd) => {
    keep_up_to_date_1.long_term_enforce_client_up_to_date(() => {
        const port = cmd.port || 6565;
        const specs_filepath = path.resolve(specsFile || 'pagedraw-specs.js');
        if (!specsFile) {
            console.warn(`No specs file specified. Trying specs file at ${specs_filepath}`);
        }
        if (cmd.config && cmd.overrideConfig) {
            console.error("Specifying -c and -r simultaneously is not supported.");
            process.exit(1);
        }
        if (!fs.existsSync(specs_filepath)) {
            console.error(`ERROR: No specs file found at ${specs_filepath}`);
        }
        else {
            const config = cmd.config || cmd.overrideConfig;
            const override = !!cmd.overrideConfig;
            ext_dev_server_1.start_dev_server(port, config, override, specs_filepath)
                .catch(() => process.exit(1));
        }
    });
});
program
    .command('*', '', { noHelp: true })
    .action(function () {
    program.outputHelp();
});
utils_1.log('Pagedraw CLI starting');
program.parse(process.argv);
if (process.argv.length <= 2) {
    program.help();
}
//# sourceMappingURL=pagedraw.js.map